#define _SYS_PARAM_H_  1
/* Hacked by Glenn Everhart 1995. Hope is to turn it into
   something that can read BSD type filesystems. Ability to still
   read V6 or V7 filesystems is NOT an issue. The intent is however
   to read "the filesystem in this collection of blocks" so
   it won't bother with partition tables. Idea is to run on systems
   with 1 partition used per disk or whose partitioning is below the
   level of this thing. */
/* Install as "gtl" to list directories, "gtf" to copy files */
/* By default uses /dev/bsd as device to list */
/* This code is intended to read BSD systems that are either big-endian
   OR little-endian in origin. This is done by reading the superblock (and
   the first super block MUST be there for this code) and looking for the
   magic number that signals it IS a superblock. If this magic number is
   not found, we assume the file structure being read is of the opposite
   "endian-ness" from whatever we're running on and for that reason do
   byte swaps and word swaps and so on as needed while we read the inodes
   and file structures. Data is NOT byteswapped however, since for textual
   files it will come across correctly in either case. We only do byte
   swapping where we need to interpret strings of 2 or 4 bytes as 16 or
   32 bit integers. In principle it is not possible to know what the "right"
   thing to do with all data is, and for cross-platform operation, doing
   nothing is the best available option. Even for network-transmitted
   binary, since that is by convention sent in "net" order, this should be
   right. */
/* Basically the ability to read V6 or V7 filesystems is ripped out. Use
   the old code for that. - gce */
/* From:uunet!SOL1.GPS.CALTECH.EDU!carl */
/*
 *	hacked jan 82 by norman wilson:
 *	- implement ls -l (lots of work; involved
 *	  removing many global variables to be able
 *	  to look at two inodes at once!)
 *	- let output file for cp be a directory (as in normal cp)
 *	- clean up some cpu dependencies
 *
 *	Lots more hacking April/May 1982 by MLB:
 *	- Added code to permit use under VMS, using the
 *	  EUNICE C compiler and LIBC.OLB
 *	- Fixed a few minor bugs
 *	- Got rid of file size limit for V6 files (but
 *	  this code has NOT been extensively tested!!!)
 */

/* AXP has "long" meaning 64 bits, but for on-disk structures, many
   "long" things are assumed to be 32 bits owing to the 32-bit origins
   of the BSD Fast File System. Therefore use type "Long" instead of
   "long" in the .H files we use and define it here. */

/* Define our 32-bit type */
#ifdef __alpha
#define __ALPHA 1
#define Long int
 typedef unsigned int u_Long;
#else
#define Long long
 typedef unsigned long u_long;
#endif
#include <stdio.h>
/* Use our own include files so we can edit them to handle idiosyncrasies
   from Sun, HP, or other barbarian tribes... */
#include "fsdir.h"
#include "fs.h"
/*#include "dinode.h"*/
#include "inode.h"
#include <sys/types.h>
#include <unistd.h>

/*
 *  Version of <sys/types.h>
 *  with Longwords set up for swapping
 *  because VAX and 11 Longs are different
 */

#if vax
typedef struct {unsigned short l_high; unsigned short l_low;} SLong;
#else
#ifndef __alpha
typedef struct {unsigned unsigned short l_low; short l_high;} SLong;
#else
typedef int SLong;
#endif
#endif vax
#define ushort unsigned short
typedef		SLong		daddr_t;
typedef		char *		caddr_t;
typedef		SLong		time_t;
typedef		SLong		paddr_t;
#include <sys/dir.h>
#include <pwd.h>
#include <sys/stat.h>
#include <time.h>
#include <string.h>

/*
#define	major(x)	(int)((unsigned)(x)>>8)
#define	minor(x)	(int)((x)&0377)
#define	makedev(x,y)	(dev_t)(((x)<<8) | (y))
*/


/* storage buffers for my reading foreign BSD file structures */
/* Note that where these are read in from bigendian systems, they
   must have various fields byteswapped and maybe wordswapped */

/* Use these to hold disk structures so we can figure what gets
   addressed where. The structures of course must be fixed up
   before they can be used. */

static struct dinode mydinode;
static struct dirent mydirent;
static struct csum mycsum;
static struct fs myfs; /* superblock */
static struct cg mycg; /*cylinder group */
short int sswapb(short int inint);

static int doswab;

/*
 *  File system/special file tables .....
 *  An entry says that filenames beginning with
 *  `pfx' are to be found on device `spcl'.
 *  NULL pfx is a catchall at end;
 *  put default (ie the root device) here.
 *
 *  For VMS version, `mindev' is the UNIX minor device
 *  number for the appropriate special file.  This is
 *  used to determine the drive number and the index
 *  into the `cyloff' table.
 */

/* buffer sizes, config for sun & dec unix */
#define	BUFSIZB	8192	/* Buffer size (bytes) */
#define	BUFSIZS	4096	/* Buffer size (shorts) */
#define	BUFSIZL	2048	/* Buffer size (Longs) */

#define	NULLP	(Long *)0	/* Uninitialized pointer */

#define	APB	BUFSIZL		/* Number of addresses per block */

union {
	char	CHARbuf[BUFSIZB];
	struct direct DIRbuf[32];
} bufu;

#define	charbuf	bufu.CHARbuf
#define	dirbuf	bufu.DIRbuf

/* Some more data areas we need to read BSD files */
static char filedat[BUFSIZB]; /* data block from file */
static int indblk[BUFSIZL]; /* indirect block, where needed */
static int nindblk; /* index of indirect block in inode re're reading */
static char dirblk[BUFSIZB]; /* directory file data block */
static int ndirblk; /* Position in directory block we're reading */
static int nsizdirb;
static int dirblkn; /* block number of the directory in dirblk */
static int filblkn; /* file inode number in filedat */
static struct direct * dirptr;
static char * cdirptr;

/* Indirect reference blocks. */

static int i2blk[BUFSIZL]; /*2nd level indirect blk*/
static int i3blk[BUFSIZL]; /* 3rd level indirect blk */
static int j2blk[BUFSIZL]; /* extra blocks needed for multiple indirection*/
static int j3blk[BUFSIZL]; /* extra blocks needed for multiple indirection*/
static int k3blk[BUFSIZL]; /* extra blocks needed for multiple indirection*/
/* Pointers to follow 1, 2, and 3-indirect pointers */
int * i2p = NULLP;
int * i3p = NULLP;
int * i1p = NULLP;

/* pointers to track file reading in getblock */
static Long ilevel,lleft;
static Long lleft2,lleft31,lleft32;

#define	EXECUTEP	01
#define	WRITEP		02
#define	READP		04

/* These don't really matter */
#define BSD	2
#define	V6	0
#define	V7	1
#define	VAX	2

#if VMS

#define	DFLTFSTYPE	BSD

#else

#if vax
#define	DFLTFSTYPE	V7
#else
#define	DFLTFSTYPE	VAX
#endif vax

#endif VMS

int	fstype = DFLTFSTYPE;

char	*progname;	/* Name of program */
char	lsflag;		/* Nonzero ==> list directory/file rather than copy file */
char	silent = 0;	/* Nonzero ==> suppress error messages */
char	Longls = 0;	/* Nonzero ==> list directory/file verbosely */
#if VMS
char	image = 0;	/* Nonzero ==> image-mode transfer */
#endif VMS
char	*device;	/* Name of device containing UNIX filesystem */
char	*inptf;		/* Name of input file */
char	*output;	/* Name of output file */
Long	offset = 0;	/* Offset value to be added value passed to lseek */
int	des;		/* File descriptor for reading input filesystem */
char	fsw;		/* Flag used to indicate end of pathname scan */
Long	bnum;		/* Number of next block to be read */
Long	iaddr;		/* Index to inode's pointer array */

#define	lastchr(x)	(x[strlen(x)-1])


main(argc,argv)
int	argc;
char	*argv[];
{
	int * intp;
	int * intp2;
	int k,kk,kkk,kkkk, ii;
	long lkk;
	char	*basename();
	char	*fullinp;
	register struct fs	*fs;
	ino_t	inum;
#if VMS
/* NOTE:  Some versions of UNIX use the name strchr() rather than index() */
	char	*index();
#endif VMS
	struct dinode	inode;

	if (argc < 2 )usage();
	doswab = 0;
	lsflag = ( *(progname=basename(*argv)) == 'l' );

	switch ( progname[2] ) {
/* does nothing useful*/
		case 'u':	fstype = BSD; break;
		case 'v':	fstype = VAX; break;
		case '1':	fstype = V7;  break;
		case 'l':	lsflag = 1;   break;
		case '6':	fstype = V6;  break;
	}
	files(argc,argv);
	fullinp = inptf;
	if ( device == NULL ) {
		device = "/dev/bsd";
	}
	if ( (des=open(device,0)) < 0 )
		error ("%s cannot be opened\n",device);
	setgid(getgid());
/* Got the device open now. Read the superblock now. */
/* Reads whole superblock into memory. */
	lseek(des, SBOFF, 0);
	if (ii =read(des, &myfs, sizeof(myfs)) == 0){
	  fprintf(stderr," Cannot read superblock\n");
	  exit(2);
	}
/* See if the superblock magic number looks OK as is. If not, bump the
   doswab variable and swap bytes in hopes that will fix it. */
/* This will attempt to handle, in this way, disks that are from bigendian
   systems AND from littleendian systems in one program. */
#define FS_MAGIC 0x11954
	if (myfs.fs_magic != FS_MAGIC) {
	   doswab = 1;
/* fix up superblock offsets we need so we can find cylinder groups */
/* swap bytes on fields */
	   intp= &myfs.fs_fmod;
	   lkk = (long)&myfs;
	   lkk = (long)intp - lkk;
	   lkk = lkk / 4;
	   for (intp = (int *)&myfs; lkk > 0; lkk--){
	      *intp = iswapb(*intp);
	      intp++;
	   }
/* Just about everything is longs here so just swap it all */
	}
/* To read files, it isn't actually necessary to read the cg structures
   at all. However we must use info in the superblock to find where they
   are. The macro cgimin(fs, c) will generate the block (actually, fragment)
   number of a given inode. Note too that dinodes (on-disk inodes) are
   found using the fact that the root inode is inode number 
   UFSROOTINO (2 generally). One reads the root inode to read the root
   directory and descends from there. Once you read an inode, there is a
   list of disk blocks and a list of indirect blocks. The file size in
   bytes is available too. Recall that the blocks here are full disk
   blocks, except fragments. The pointers however are expressed as
   fragment numbers. The fragments are 1024 byte chunks of disk.
   Large files are all fullsize blocks, but small ones can use only
   partial blocks. */

/* Now read the root inode */
	inum = UFSROOTINO;
        iread(&mydinode, inum);

/* The inode contains a few direct data block pointers and 3 indirect
   blocks. Nothing seems to directly indicate how these are used, but
   considerations of size would seem to indicate that they are singly,
   doubly, and trebly indirect blocks. If they were all singly indirect
   each would hold 2048 longwords which could refer to a disk block
   each, and 6000 blocks, even at 8kb each, is only around 49 megs.
   With 8 bytes reserved for file size, this doesn't make much sense.
   If the blocks have doubly and trebly indirect pointers, however,
   file sizes can get huge and things are more sane. */

	while ( *inptf && *inptf == '/' )
		inptf++;
	fsw = *inptf ? 1 : 0;
	while ( fsw ) {
		iread(&inode,inum);
		access(&inode,EXECUTEP);
		if ( (inum=trace(&inode)) == 0)
			error("%s -- input file not found\n",fullinp);
	}
/* get the dir or file we want now by getting its inode */
	iread(&inode,inum);
	if ( lsflag )
		dols(fullinp,&inode);
	else
		docp(fullinp,&inode,output);
	exit(0);
}


files(ac,av)
int	ac;
char	**av;
{
	int	dflg = 0;

	if (av != NULL){
	if (*av != NULL && ac > 1){
	 while ( **++av == '-' ) {
		--ac;
		while ( *++*av ) {
			switch ( **av ) {
			case 's':	++silent; break;
			case 'l':	++Longls; ++lsflag; break;
			case 'p':	++lsflag; break;
			case '6':	fstype = V6; break;
			case 'c':	lsflag = 1; break;
			case '7':	fstype = V7; break;
			case 'v':	fstype = VAX; break;
			case 'u':	fstype = BSD; break;
#if VMS
			case 'i':	++image; break;
#else
			case 'd':	if ( ac < 3 )
						usage();
					++dflg; break;
#endif VMS
			default:	error("Illegal flag (%c)\n",**av);
			}
		}
		if ( dflg ) {		/* Kludge, sort of */
			device = *++av;
			--ac;
			dflg = 0;
		}
	 }
	}
	}
	inptf = *av;
	if ( !lsflag )
		output = (ac>2) ? *++av : NULL;
}


usage()
{
	error("Usage: %s [-u] [-o] [-n] [-d device] input-file %s\n",
		progname, lsflag? "" : "[output-file]");
}


chkfs(av1,fs)
char	*av1;
char	*fs;
{
	int	len;

	len = strlen(fs);
	if ( ( strncmp(av1,fs,len--) == 0 )
	|| ( ( av1[len] == '\0' ) && ( strncmp(av1,fs,len) == 0 ) ) ) {
		inptf += len;
		return (1);
	}
	return (0);
}


trace(ip)
struct dinode	*ip;
{
	char	*getname();
	char	*name;
	int kkk;
	ino_t	inum;

	name = getname();
	nsizdirb= isize(ip);
	for ( bnum=0; getblock(ip,&dirblk[0]); ){
		if ( (inum=dentry(name)) != 0 )
			return(inum);
	kkk = BUFSIZB;
	if (nsizdirb < kkk) kkk = nsizdirb;
	nsizdirb = nsizdirb - kkk;
	if(kkk <= 0)break;		
	}
	return(0);
}


char *
getname()
{
#define DIRSIZ 1024
	register int	j;
	static char	name[DIRSIZ+1];

	for ( j=0; j<DIRSIZ && *inptf!='/'; j++ ) {
		if ( *inptf == '\0' )
			break;
		name[j] = *inptf++;
	}
	name[j] = '\0';
	while ( *inptf && *inptf != '/' )
		inptf++;
	while ( *inptf && *inptf == '/' )
		inptf++;
	if ( *inptf == '\0' )
		fsw = 0;
	return (name);
}


dentry(name)
char	*name;
{
	register int	m;
	ino_t		inum;
	int kk,kkk, inowk, nbytes, dbsiz;
/* Note we swap bytes as needed while reading directories here. */

	dbsiz = BUFSIZB;
	if (nsizdirb < dbsiz) dbsiz = nsizdirb;
	kk = myfs.fs_bsize / 8L;
/* 8L is minimum size of a directory entry */
/* (actual useful min is 12) */
	(char *)dirptr = &dirblk[0];
	for ( m=0; m<kk; m++ ) {
/* track thru variable size directory entries to look for name */
		if ( (inum=dirptr->d_ino) == 0){
			  kkk = dirptr->d_reclen;
			  if (doswab != 0) kkk = sswapb(kkk);
			  (long)dirptr = (long)dirptr + kkk;
			  continue;
			}
		if (doswab != 0){
/* Must swap bytes of inum */
		  inum = iswapb(inum);
		}
	   kkk = dirptr->d_namlen;
	   if (doswab != 0) kkk = sswapb(kkk);
	   if ( strncmp(&dirptr->d_name, name, kkk) == 0){
			  return(inum);
	   }
	  kkk = dirptr->d_reclen;
	  if (doswab != 0) kkk = sswapb(kkk);
	  (long)dirptr = (long)dirptr + kkk;
	  if (((long)dirptr >= (long)(&dirblk[0] + dbsiz))) break;
	}
	return(0);
}


dols(fullname,ip)
char	*fullname;
register struct dinode	*ip;
{
	Long		isize();
	Long		nbytes;
	register int	k;
	int kk,kkk,kkkk,m,dbsiz,iino;
	int idun;
	long kuu;
	ino_t	inum;
	Long		size = 0;
	char		fn[DIRSIZ+1];

	if ( (ip->di_mode & S_IFMT) != S_IFDIR ){
		lsone(fullname,ip,0);
	} else {
		access(ip,READP);
		redirout();
		bnum = 0;
		nbytes = isize(ip);

	kk = BUFSIZB/12;
	idun = isize(ip);
	nsizdirb= isize(ip);
	while (getblock(ip,&dirblk[0]) && idun > 0){
	dbsiz = BUFSIZB;
	if (idun < dbsiz) dbsiz = idun;
	idun = idun - dbsiz;
	(char *)dirptr = &dirblk[0];
	for ( m=0; m<kk; m++ ) {
/* track thru variable size directory entries to look for name */
		if ( (inum=dirptr->d_ino) == 0){
			  kkk = dirptr->d_reclen;
			  if (doswab != 0) kkk = sswapb(kkk);
			  (long)dirptr = (long)dirptr + kkk;
			  continue;
			}
		if (doswab != 0){
/* Must swap bytes of inum */
		  inum = iswapb(inum);
		}
	   kkk = dirptr->d_namlen;
	   if (doswab != 0) kkk = sswapb(kkk);
	   if(dirptr->d_ino != 0){
#ifdef ERDIAG
	fprintf(stderr,"direntry %s\n",&dirptr->d_name[0]);
#endif
	        strncpy(fn,&dirptr->d_name[0],DIRSIZ);
			iino = dirptr->d_ino;
			if (doswab != 0)iino=iswapb(iino);
			lsone(fn,NULL,iino);
	   }
	  kkk = dirptr->d_reclen;
	  if (doswab != 0) kkk = sswapb(kkk);
	  (long)dirptr = (long)dirptr + kkk;
#ifdef ERDIAG
	kuu = (long)(&dirblk[0]) + dbsiz;
	fprintf(stderr,"dirptr,end %d  %d\n",(long)dirptr,kuu);
#endif
	  if (((long)dirptr >= (long)(&dirblk[0] + dbsiz))) break;
	}

	}
		fclose(stdout);
		wait(0);
	}
}


lsone(name,ip,inum)
char	*name;
struct dinode	*ip;
ino_t	inum;
{
	static struct dinode	inode;
	int	i_uid;
	short	i_mode;
	short	i_nlink;
	int	li_dev;
	Long	i_mtime;
	Long	fixl();
	char	*uidname();
	Long	isize();
	char	*tstr();
	char	li_type;

	if ( Longls == 0 ) {
		printf("%s\n",name);
		return;
	}
	if ( ip == NULL ) {
		ip = &inode;
		iread(ip,inum);
	}
		i_uid   = ip->di_uid;
		i_mode  = ip->di_mode;
		i_nlink = ip->di_nlink;
		li_dev = 1;
		i_mtime = ip->di_mtime;
	putchar(li_type=ftype(i_mode));
	pmode(i_mode);
	printf("%2d %-6.6s",i_nlink,uidname(i_uid));
	if ( li_type == 'b' || li_type == 'c' )
		printf("%3d,%3d",1,1);
	else
		printf("%7ld",isize(ip));
	printf(" %s %s\n",tstr(i_mtime),name);
}


iread(ip,ino)
/*struct dinode	*ip;*/
struct icommon *ip;
register ino_t	ino;
{
	Long	j;
	int kk,kkk,inodg,inodd;
	long xkkk,xkkkk;
	struct fs * fsp;

	xkkk = (long)ip;
	fsp = &myfs;
/* Form cylinder number in which inode is located */
	inodg = ino / fsp->fs_ipg;
	inodd = (ino - (inodg * fsp->fs_ipg));
/* Now find the block in which this inode lives */
	j = cgimin(fsp, inodg);
/* Be sure and adjust INODSIZ to be the size of an inode if it
   isn't 128 bytes. */
#define INODSIZ 128
	if ( lseek(des,myfs.fs_fsize*j+inodd*INODSIZ+offset,0) == -1 ) {
		perror("lseek");
		error("I/O error\n");
	}
/* Read in the inode */
	if ( read(des,ip,sizeof(mydinode)) < 0 ) {
		perror("read");
		error("I/O error\n");
	}

/* Now if the read was OK, maybe swap bytes */
	if (doswab != 0){
/* gotta swap things around */
	   ip->ic_smode = sswapb(ip->ic_smode);
	   ip->ic_nlink = sswapb(ip->ic_nlink);
	   ip->ic_suid = sswapb(ip->ic_suid);
	   ip->ic_size = iswapb(ip->ic_size);
	   ip->ic_size2 = iswapb(ip->ic_size2); /* low order size */
	   ip->ic_atime = iswapb(ip->ic_atime);
	   ip->ic_mtime = iswapb(ip->ic_mtime);
	   ip->ic_ctime = iswapb(ip->ic_ctime);
	   for (kk = 0; kk < NDADDR; kk++){
		ip->ic_db[kk] = iswapb(ip->ic_db[kk]);
	   }
	   for (kk = 0; kk < NIADDR; kk++){
		ip->ic_ib[kk] = iswapb(ip->ic_ib[kk]);
	   }
	   ip->ic_flags = iswapb(ip->ic_flags);
	   ip->ic_blocks = iswapb(ip->ic_blocks);
	   ip->ic_gen = iswapb(ip->ic_gen);
	   ip->ic_shadow = iswapb(ip->ic_shadow);
	   ip->ic_uid = iswapb(ip->ic_uid);
	   ip->ic_gid = iswapb(ip->ic_gid);
	}
	xkkkk = (long)&mydinode;
	if(xkkk == xkkkk){
/* Now read the indirect blocks in so they'll BE there */
/* However only read them if they're present. */
/* zero the indirect blocks first to prevent problems from old data*/
	for(kk=0;kk < BUFSIZL; kk++){
	  indblk[kk] = 0L;
	  i2blk[kk] = 0L;
	  i3blk[kk] = 0L;
	}
/* init the pointers when we refill buffers */
	i1p = NULL;
	i2p = NULL;
	i3p = NULL;
	if(ip->ic_ib[0] != 0)get(ip->ic_ib[0],&indblk[0]);
	if(ip->ic_ib[1] != 0)get(ip->ic_ib[1],&i2blk[0]);
	if(ip->ic_ib[2] != 0)get(ip->ic_ib[2],&i3blk[0]);
	}
}


ftype(mode)
register ushort	mode;
{
	register ushort	filtyp;

	filtyp = mode & S_IFMT;

	if ( filtyp == S_IFREG )
		return('-');
	if ( filtyp == S_IFDIR )
		return('d');
	if ( filtyp == S_IFBLK )
		return('b');
	if ( filtyp == S_IFCHR )
		return('c');
	if ( filtyp == S_IFIFO )
		return('p');
	return('?');
}


/*
 *  Code to print the mode symbolically.
 *  Originally stolen straight from ls.
 *  Rewritten to put the program in the public domain,
 *  which is a pretty stupid waste of time
 *  (the rewriting), but less so than SMP.
 */

char mchars[] = "rwxrwxrwx";

pmode(mode)
register int	mode;
{
	char		buf[sizeof(mchars)];
	register int	bit;
	register char	*s;

	strcpy(buf,mchars);
	if ( mode & S_ISVTX ) {
		buf[8] = 't';
		mode |= S_IEXEC >> 6;
	}
	if ( mode & S_ISGID ) {
		buf[5] = 's';
		mode |= S_IEXEC >> 3;
	}
	if ( mode & S_ISUID ) {
		buf[2] = 's';
		mode |= S_IEXEC;
	}
	for ( bit = S_IREAD, s = buf; bit; bit >>= 1, s++ )
		if ((mode & bit) == 0)
			*s = '-';
	printf("%s",buf);
}


char *
uidname(uid)
int	uid;
{
	struct passwd	*pw;
	static char	buf[10];
	struct passwd	*getpwuid();

	if ( (pw=getpwuid(uid)) != NULL )
		return (pw->pw_name);
	else {
		sprintf(buf,"%d",uid);
		return (buf);
	}
}


char *
tstr(t)
Long t;
{
	Long		time();
	static Long	year;
	char		*ctime();
	static char	buf[15];
	register char	*cp;

	if ( year == 0 )
		year = time(0) - 60L*60L*24L*30L*6L;	/* 6 months ago */
	cp = ctime(&t);

	if ( year < t && t < time(0) )
		sprintf(buf,"%-12.12s",cp+4);
	else
		sprintf(buf,"%-7.7s %-4.4s",cp+4,cp+20);
	return (buf);
}


redirout()
{
	int	pfd[2];
	int	fk;

	return;
/*
#if VMS
	return;
#else
	if ( pipe(pfd) )
		error("Can't create pipe\n");
	while ( (fk=fork()) == -1 )
		sleep(1);
	if ( fk == 0 )
		dup2(pfd[0],0);
	else
		dup2(pfd[1],1);
	close(pfd[0]);
	close(pfd[1]);
	if ( fk == 0 ) {
		execl("/usr/bin/sort","sort",Longls ? "+0.40" : 0, 0);
		execl("/bin/sort","sort",Longls ? "+0.40" : 0, 0);
		error("Can't exec sort\n");
	}
#endif VMS
*/
}


docp(fullname,ip,outfile)
char	*fullname;
register struct dinode	*ip;
char	*outfile;
{
	int	odes;
	Long	isize();
	Long	size = 0L;
	Long	nbytes;
	int	k;

	if ( (ip->di_mode & S_IFMT) == S_IFDIR )
		error("%s is a directory\n",fullname);
	if ( (ip->di_mode & S_IFMT) != S_IFREG )
		error("%s is a special file\n",fullname);
	else {
		access(ip,READP);
		odes = openout(outfile,fullname);
		bnum = 0;
		nbytes = isize(ip);
		while ( getblock(ip,&charbuf[0]) ) {
			k = BUFSIZB;
			if ( size+BUFSIZB > nbytes )
				k = nbytes - size;
			write(odes,charbuf,k);
			size += k;
		}
	}
	close(odes);
}


#define OWNER	6
#define GROUP	3
#define GENRL	0

#if VMS
#define	SU_UID	4
#define	SU_GID	1
#else
#define SUSER	0
#endif

access(ip,key)
register struct dinode	*ip;
register int	key;
{
	int	uid;
	int	i_uid;
	int	i_gid;
	int	i_mode;

	/* superuser has all permissions */
		return;
}


openout(outname,inname)
char *outname;
char *inname;
{
	char		c;
	char		*outnm;
	char		*basename();
	int		odes;
	static char	buf[80];

	if ( outname == NULL )
		return (fileno(stdout));
#if VMS
	outnm = outname;
	if ( (c=lastchr(outname)) == ':' || c == ']' ) {
		strcpy(buf,outname);
		strcat(buf,basename(inname));
		outnm = buf;
	}
	if ( (odes=(image?creat(outnm,0644):creat(outnm,0644,"txt"))) < 0 )
		error("cannot creat %s\n",outname);
#else
	if ( (odes=creat(outname,0666)) < 0 ) {
		sprintf(buf,"%s/%s",outname,basename(inname));
		if ( (odes=creat(buf,0666)) < 0 )
			error("cannot creat %s\n",outname);
	}
#endif VMS
	return (odes);
}


char *
basename(s)
register char	*s;
{
/* NOTE:  Some versions of UNIX use the name strrchr() rather than rindex() */
	char		*rindex(char* a, int b);
	register char	*t;

	if ( (t=rindex(s,'/')) == NULL )
		return(s);
	else
		return(t+1);
}
char *
rindex(s,t)
char *s;
int t;
{
  return (strrchr(s,t));
}

/*
 * Getblock -- return next block in file
 *
 *	Does all indirection, etc.  BUT doesn't look
 *	at length of file, so it always returns a
 *	full block.  The other code has to watch out.
 */

/* NDADDR is number of directly addressed blocks (see inode.h) */

/* Presume (for lack of better info) that disk addresses are packed
   as 4-byte structures */
getblock(ip,bufptr)
struct icommon	*ip;
char* bufptr;
{
	Long		isize();
	static Long	nbytes;
	Long		block;
	char	*p;
	int * inaddrp;

	if ( bnum == 0 ) {	/* new file */
		iaddr = 0;
		nbytes = isize(ip);
		ilevel = 0;
		lleft2=0;
		lleft31=0;
		lleft32=0;
		lleft = 0;
		inaddrp = NULL;
	}
	if ( myfs.fs_fsize*bnum >= nbytes || nbytes == 0 )
		return(0);	/* end of file */

/* do 4 bytes at a time, not 3, if bsd uses 4 */
/* I can't believe packing would be 3 bytes only. Use 4 */
/* This way of forming block number is correct whether doswab is set or not*/
/* Get the block number among indirect blocks first.*/

/* the blocks in the file are as follows:
   0-NDADDR - 1 		: in inode ic_db array
   NDADDR - NDADDR + BUFSIZL    : in ic_ib[0] block are blk nos.
   NDADDR + BUFSIZL - NDADDR + BUFSIZL +BUFSIZL^2 : in ic_ib[1] blk
...similar for ic_ib[2] for really huge files.
*/
		if ( bnum < NDADDR ) {
			block = ip->ic_db[bnum];
			get(block,bufptr);
			++iaddr;
			++bnum;
			return(1);
		}
		if (inaddrp = NULL){
		  inaddrp = indblk;
		  ilevel = 1;
		  lleft = BUFSIZL;
		}
		p = (char *)inaddrp;
		block =  (((Long)p[0])&0377)
		      + ((((Long)p[1])&0377)<<8)
		      + ((((Long)p[2])&0377)<<16)
		      + ((((Long)p[3])&0377)<<24);

		get(inaddrp,bufptr);
/* update pointers etc. if we need to for up to 2 ind. levels */
		if (ilevel == 1){
		  if (lleft > 0){
		    lleft--;
		    inaddrp++;
		  }else{
		    ilevel = 2;
		    lleft = BUFSIZL;
/* indirect 2nd level gets from 2 level indirect area */
		    get(i2blk[0],j2blk);
		    inaddrp = j2blk - 1;
		    lleft2 = BUFSIZL;
		  }
		}
		if (ilevel == 2){
		  if (lleft > 0){
		    if (lleft2 > 0){
		      lleft2--;
		      inaddrp++;
		    }else{
		      lleft--;
		      get(i2blk[BUFSIZL - lleft],j2blk);
		      inaddrp = j2blk;
		      lleft2 = BUFSIZL;
		    }
		  }
		  if (lleft <= 0){
/* out of doubly indirect blocks. Go to trebly. */
		      lleft = BUFSIZL;
		      ilevel = 3;
		      lleft31 = BUFSIZL;
		      lleft32 = BUFSIZL;
		      get(i3blk[0],j3blk);
		      get(j3blk[0],k3blk);
		      inaddrp = k3blk - 1;
		  }
		}
		if (ilevel == 3){
		  if(lleft31 > 0){
		    if(lleft32 > 0){
			inaddrp++;
			lleft32--;
		    }else{
			lleft31--;
			lleft32=BUFSIZL;
			get(j3blk[BUFSIZL - lleft31], k3blk);
			inaddrp = k3blk;
		    }
		  }
		  if (lleft31 <= 0){
		    lleft--;
		    get(i3blk[BUFSIZL - lleft],j3blk);
		    get(j3blk[0],k3blk);
		    lleft31 = BUFSIZL;
		    lleft32 = BUFSIZL;
		    inaddrp = k3blk;
		  }
/* don't worry about running off the end of trebly indirect blocks */
		}

		++bnum;
		return(1);
}


getfib(block,buf)
Long	block;
char	*buf;
{
	getib(block,buf);
	++iaddr;
}


getib(block,buf)
Long	block;
char	*buf;
{
	get(block,buf);
/*	swaw(buf); */
}


swaw(buf)
int	*buf;
{
	register int	*p;
	register int	i;

	if(doswab != 0){
	  for ( p=buf,i=0; i<BUFSIZL; ++i ) {

		i = iswapb(*p);
		*p = i;
		p++;
	  }
	}
}


/*
 *  Get a block.  If block is 0, it is a hole in the file; act appropriately.
 */

get(block,buf)
Long	block;
char	*buf;
{
	register char *p;

	if ( block == 0 ) {
		fprintf(stderr,"%s: file has hole(s)\n",progname);
		for ( p=buf; p<buf+BUFSIZB; )
			*p++ = 0;
	} else {
/* Use the fact that we read the superblock into myfs structure */
		if ( lseek(des,myfs.fs_fsize*block+offset,0) == -1 ) {
			perror("lseek");
			error("I/O error\n");
		}
		if ( read(des,buf,BUFSIZB) != BUFSIZB ) {
			perror("read");
			error("IO error\n");
		}
	}
}


Long
isize(ip)
register struct icommon	*ip;
{
	Long	size;
	Long	fixl();
/* I believe low order (32 bit) size is in size2 */
/* (at least from sun ) */
	if (doswab != 0){
	size = ip->ic_size2;
	} else{
	size = ip->ic_size;
	}
	return (size);
}

/* VARARGS1 */

error(string,arg1,arg2)
char	*string;
char	*arg1;
char	*arg2;
{
	if ( !silent )
		fprintf(stderr,string,arg1,arg2);
	exit(-1);
}
int swapw(int inint)
{
  unsigned short * pc;
  unsigned short * pco;
  int ival, oval;
/* swap words but not bytes */
  ival = inint;
  pc = (unsigned short *)&ival;
  pco = (unsigned short *) &oval;
  *pco = *(pc+2);
  *(pco+2) = *pc;
  return(oval);
}

/* byteswap functions. Swap bytes in 4 byte units or 2 byte units. */
int iswapb(int inint)
{
  unsigned char *pc;
  unsigned char *pco;
  int ival,oval;
  ival = inint;
  pc = (unsigned char *)&ival;
  pco = (unsigned char *)&oval;
  pco = pco + 3L;
  *pco = *pc;
  pco--;
  pc++;
  *pco = *pc;
  pco--;
  pc++;
  *pco = *pc;
  pco--;
  pc++;
  *pco = *pc;
  return(oval);
}
short int sswapb(short int inint)
{
  unsigned char *pc;
  unsigned char *pco;
  short int ival,oval;
  ival = inint;
  pc = (unsigned char *)&ival;
  pco = (unsigned char *)&oval;
  pco = pco + 1L;
  *pco = *pc;
  pco--;
  pc++;
  *pco = *pc;
  return(oval);
}
